/*
 * Copyright (C) 2011-2022 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include "qgs_server.h"
#include "qgs_log.h"
#include <iostream>
#include <fstream>
#include <linux/vm_sockets.h>

#define QGS_CONFIG_FILE "/etc/qgs.conf"
#define QGS_UNIX_SOCKET_FILE "/var/run/tdx-qgs/qgs.socket"

using namespace std;
using namespace intel::sgx::dcap::qgs;
volatile bool reload = false;
static QgsServer* server = NULL;

void signal_handler(int sig)
{
    switch(sig)
    {
        case SIGTERM:
            if (server)
            {
                reload = false;
                server->shutdown();
            }
            break;
        case SIGHUP:
            if (server)
            {
                reload = true;
                server->shutdown();
            }
            break;
        default:
            break;
    }
}


int main(int argc, const char* argv[])
{
    bool no_daemon = false;
    unsigned long int port = 0;
    unsigned long int num_threads = 0;
    unsigned long int mode = 0;
    char *endptr = NULL;
    if (argc > 4) {
        cout << "Usage: " << argv[0] << " [--no-daemon] [-p=port_number] [-m=unix_socket_mode] [-n=number_threads] [--verbose] [--debug]"
             << endl;
        exit(1);
    }

    for (int i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "--no-daemon") == 0) {
            cout << "--no-daemon option found, will not run as a daemon"
                 << endl;
            no_daemon = true;
            continue;
        } else if (strcmp(argv[i], "--debug") == 0) {
            qgs_verbose = qgs_debug = true;
            continue;
        } else if (strcmp(argv[i], "--verbose") == 0) {
            qgs_verbose = true;
            continue;
        } else if (strncmp(argv[i], "-p=", 3 ) == 0) {
            if (strspn(argv[i] + 3, "0123456789") != strlen(argv[i] + 3)) {
                cout << "Please input valid port number" << endl;
                exit(1);
            }
            errno = 0;
            port = strtoul(argv[i] + 3, &endptr, 10);
            if (errno || strlen(endptr) || (port > UINT_MAX) ) {
                cout << "Please input valid port number" << endl;
                exit(1);
            }
            cout << "port number [" << port << "] found in cmdline" << endl;
            continue;
        } else if (strncmp(argv[i], "-m=", 3 ) == 0) {
            if (strspn(argv[i] + 3, "0123456789") != strlen(argv[i] + 3)) {
                cout << "Please input valid socket mode" << endl;
                exit(1);
            }
            errno = 0;
            mode = strtoul(argv[i] + 3, &endptr, 8);
            if (errno || strlen(endptr) || (mode > UINT_MAX) ) {
                cout << "Please input valid socket mode" << endl;
                exit(1);
            }
            cout << "socket mode [" << oct << mode << dec << "] found in cmdline" << endl;
            continue;
        } else if (strncmp(argv[i], "-n=", 3) == 0) {
            if (strspn(argv[i] + 3, "0123456789") != strlen(argv[i] + 3)) {
                cout << "Please input valid thread number" << endl;
                exit(1);
            }
            errno = 0;
            num_threads = strtoul(argv[i] + 3, &endptr, 10);
            if (errno || strlen(endptr) || (num_threads > 255)) {
                cout << "Please input valid thread number[0, 255]" << endl;
                exit(1);
            }
            cout << "thread number [" << num_threads << "] found in cmdline" << endl;
            continue;
        } else {
            cout << "Usage: " << argv[0] << " [--no-daemon] [-p=port_number] [-m=unix_socket_mode] [-n=number_threads] [--verbose] [--debug]"
                << endl;
            exit(1);
        }
    }


    // Use the port number in QGS_CONFIG_FILE if no valid port number on
    // command line
    if (port == 0 || num_threads == 0 || mode == 0) {
        ifstream config_file(QGS_CONFIG_FILE);
        if (config_file.is_open()) {
            string line;
            while(getline(config_file, line)) {
                line.erase(remove_if(line.begin(), line.end(), ::isspace),
                        line.end());
                if(line.empty() || line[0] == '#' ) {
                    continue;
                }
                auto delimiterPos = line.find("=");
                if (delimiterPos == std::string::npos) {
                    continue;
                }
                auto name = line.substr(0, delimiterPos);
                if (name.empty()) {
                    cout << "Wrong config format in " << QGS_CONFIG_FILE
                         << endl;
                    exit(1);
                }
                char value[256] = {};
                strncpy(value, line.substr(delimiterPos + 1).c_str(), sizeof(value) - 1);
                value[255] = '\0';
                if (!port && name.compare("port") == 0) {
                    errno = 0;
                    endptr = NULL;
                    port = strtoul(value, &endptr, 10);
                    if (errno || strlen(endptr) || (port > UINT_MAX) ) {
                        cout << "Please input valid port number in "
                             << QGS_CONFIG_FILE << endl;
                        exit(1);
                    }
                } else if (!mode && name.compare("socket_mode") == 0) {
                    errno = 0;
                    endptr = NULL;
                    mode = strtoul(value, &endptr, 8);
                    if (errno || strlen(endptr) || (mode > UINT_MAX)) {
                        cout << "Please input valid socket mode in "
                             << QGS_CONFIG_FILE << endl;
                        exit(1);
                    }
                } else if (!num_threads && name.compare("number_threads") == 0) {
                    errno = 0;
                    endptr = NULL;
                    num_threads = strtoul(value, &endptr, 10);
                    if (errno || strlen(endptr) || (num_threads > 255)) {
                        cout << "Please input valid thread number[0, 255] in "
                             << QGS_CONFIG_FILE << endl;
                        exit(1);
                    }
                }
                // ignore unrecognized configs.
            }
        } else {
          cout << "Failed to open config file " << QGS_CONFIG_FILE << endl;
        }
    }

    if (port == 0) {
        cout << "Use unix socket: " << QGS_UNIX_SOCKET_FILE << endl;
    }

    if(!no_daemon && daemon(0, 0) < 0) {
        exit(1);
    }

    QGS_LOG_INIT_EX(no_daemon);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGHUP, signal_handler);
    signal(SIGTERM, signal_handler);
    QGS_LOG_INFO("Added signal handler\n");

    try {
        do {
            reload = false;
            asio::io_service io_service;
            gs::endpoint ep;
            if (port) {
                struct sockaddr_vm vm_addr = {};
                vm_addr.svm_family = AF_VSOCK;
                vm_addr.svm_reserved1 = 0;
                vm_addr.svm_port = port & UINT_MAX;
                vm_addr.svm_cid = VMADDR_CID_ANY;
                asio::generic::stream_protocol::endpoint vsock_ep(&vm_addr, sizeof(vm_addr));
                ep = vsock_ep;
            } else {
                asio::local::stream_protocol::endpoint unix_ep(QGS_UNIX_SOCKET_FILE);
                ep = unix_ep;
            }
            QGS_LOG_INFO("About to create QgsServer with num_thread = %d\n", (uint8_t)num_threads);
            server = new QgsServer(io_service, ep, (uint8_t)num_threads);
            /* Allow mode to be determined by umask by default,
             * overriding only if an explicit mode is requested
             */
            if (!port && mode != 0) {
                chmod(QGS_UNIX_SOCKET_FILE, mode);
            }
            QGS_LOG_INFO("About to start main loop\n");
            io_service.run();
            QGS_LOG_INFO("Quit main loop\n");
            QgsServer *temp_server = server;
            server = NULL;
            QGS_LOG_INFO("Deleted QgsServer object\n");
            delete temp_server;
        } while (reload == true);
    } catch (std::exception &e) {
        cerr << e.what() << endl;
        exit(1);
    }

    QGS_LOG_FINI();
    return 0;
}
